/*******************************************************************************
 * Copyright (c) 2010 Neil Bartlett.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Neil Bartlett - initial API and implementation
 ******************************************************************************/
package bndtools.runtime.eclipse.applaunch;

import java.util.concurrent.Executor;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.osgi.service.runnable.ApplicationLauncher;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.BundleException;
import org.osgi.framework.Constants;
import org.osgi.framework.Filter;
import org.osgi.framework.FrameworkUtil;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceRegistration;
import org.osgi.util.tracker.BundleTracker;
import org.osgi.util.tracker.ServiceTracker;

public class Activator implements BundleActivator {

    final Logger log = Logger.getLogger(Activator.class.getPackage().getName());

	private MainThreadExecutorTracker executorTracker;
    private BundleTracker bundleTracker;

	public void start(BundleContext context) throws Exception {
		// Track the main-thread executor and register the Eclipse ApplicationLauncher around it
		executorTracker = new MainThreadExecutorTracker(context);
		executorTracker.open();

		// Ensure that the Equinox Application Admin bundle is fully started
		bundleTracker = new BundleTracker(context, Bundle.INSTALLED | Bundle.RESOLVED | Bundle.STARTING, null) {
		    @Override
		    public Object addingBundle(Bundle bundle, BundleEvent event) {
		        if("org.eclipse.equinox.app".equals(bundle.getSymbolicName())) {
		            try {
                        bundle.start();
                    } catch (BundleException e) {
                        log.log(Level.WARNING, "Error starting org.eclipse.equinox.app bundle", e);
                    }
		        }
		        return bundle;
		    }
		};
		bundleTracker.open();
	}

	public void stop(BundleContext context) throws Exception {
	    bundleTracker.close();
		executorTracker.close();
	}
}

class MainThreadExecutorTracker extends ServiceTracker {

	final Logger log = Logger.getLogger(MainThreadExecutorTracker.class.getPackage().getName());

	MainThreadExecutorTracker(BundleContext context) {
		super(context, createFilter(), null);
	}
	static Filter createFilter() {
		try {
			String filterStr = String.format("(&(%s=%s)(thread=main))",
					Constants.OBJECTCLASS, Executor.class.getName());
			return FrameworkUtil.createFilter(filterStr);
		} catch (InvalidSyntaxException e) {
			throw new RuntimeException(e);
		}
	}
	@Override
	public void open(boolean trackAllServices) {
		log.log(Level.FINE, "Starting to track main-thread executor service. Filter = {0}.", filter);
		super.open(trackAllServices);
	}
	@Override
	public void close() {
		super.close();
		log.fine("Stopped tracking main-thread executor service.");
	}
	@Override
	public Object addingService(ServiceReference reference) {
		Executor executor = (Executor) context.getService(reference);
		log.fine("Got instance of main-thread executor service -- registering ApplicationLauncher service.");
		ServiceRegistration reg = context.registerService(ApplicationLauncher.class.getName(), new ExecutorApplicationLauncher(executor, context), null);
		return reg;
	}
	@Override
	public void removedService(ServiceReference reference, Object service) {
		ServiceRegistration reg = (ServiceRegistration) service;
		log.fine("Main-thread executor service going away -- unregistering ApplicationLauncher service.");
		reg.unregister();

		context.ungetService(reference);
	}
}

class NewThreadExecutor implements Executor {
	public void execute(Runnable command) {
		new Thread(command).start();
	}
}
